
/* Generic Queue manipulation routines 
   NOTE: All elements of Queue should have QUEUE_CHAIN
   as the first element
*/

 

/*{{{  QEmpty*/
/*	Is the queue empty ? 
*/
#define QEmpty( q )		( ((void *)q) == NULL )
/*}}}*/

/*{{{  InitQ*/
/*	Initialize a queue
*/
#define	InitQ( q )		q = NULL
/*}}}*/

/*{{{  EnQueue*/
/* 	Add 'element' to the tail of the specified queue
*/
#define EnQueue( q, element )	{					\
	((QUEUE_CHAIN *)element)->Next = NULL;				\
	if ( QEmpty(q) ) q = element;					\
	else {								\
		QUEUE_CHAIN *temp;					\
		temp = (QUEUE_CHAIN *)q;				\
		while ( temp->Next != NULL )				\
			temp = temp->Next;				\
		temp->Next = (QUEUE_CHAIN *)element;			\
	}								\
}
/*}}}*/

/*{{{  DeQueue*/
/* 	Obtain an element from the head of the specified queue
*/
#define DeQueue( q, element )	{					\
	if ( QEmpty(q) ) element = NULL;				\
	else {								\
		element = q;						\
		q = ((QUEUE_CHAIN *)q)->Next;				\
	}								\
}
/*}}}*/

/*{{{  DeQueueN*/
/* 	Obtain N elements from the head of the specified queue
*/
#define DeQueueN( q, n, list )	{					\
	int i;								\
	QUEUE_CHAIN *last;						\
	list = q;							\
	for ( i = 0; ((i < n)&&(q != NULL)); i++,  q = ((QUEUE_CHAIN *)q)->Next ) last = (QUEUE_CHAIN *)q;\
	if ( i != n ) {							\
		q = list; list = NULL;					\
	} else								\
		((QUEUE_CHAIN *)last)->Next = NULL;			\
}
/*}}}*/
/*{{{  QHead*/
/* 	Return the element at the head of the specified queue
*/
#define QHead( q ) ( q )

/*}}}*/

/*{{{  QNext*/
/*	Return the element after the specified element in the queue
*/
#define	QNext( element ) ((element == NULL)?NULL:			\
				((QUEUE_CHAIN *)element)->Next)
/*}}}*/

/*{{{  QRemove*/
/*	Remove the specified element from the queue. Do not free the element.
*/
#define QRemove( q, element ) {						\
	QUEUE_CHAIN *temp = q;						\
	if ( q == element )						\
		q = ((QUEUE_CHAIN *)element)->Next;			\
	else {								\
		while (( temp != NULL) && ( temp->Next!= (void *)element ))	\
			temp = temp->Next;				\
		if ( temp != NULL )					\
			temp->Next = temp->Next->Next;			\
	}								\
}
/*}}}*/

/*{{{  QDelete*/
/*	Delete the specified element from the queue and free the memory
 	allocated to it.
*/
#define QDelete( q, element ) {						\
	QRemove( q, element );						\
	free( element );						\
}
/*}}}*/

/*{{{  InsertQ*/
/*	Add an element to the Queue after 'prev'
*/
#define InsertQ( q, element, prev ){					\
	if ( q == NULL ) { q = element; 				\
		((QUEUE_CHAIN *)element)->Next = NULL; }		\
	else if ( prev == NULL ) { 					\
		((QUEUE_CHAIN *)element)->Next = q; 			\
		q = element; 						\
	}								\
	else {								\
		((QUEUE_CHAIN *)element)->Next = ((QUEUE_CHAIN *)prev)->Next;\
		((QUEUE_CHAIN *)prev)->Next = element;			\
	}								\
}
/*}}}*/

/*{{{  AddList*/
/* 	Add 'element' to the head of the specified queue
*/
#define AddList( q, element )	{					\
	((QUEUE_CHAIN *)element)->Next = q;				\
	q = element;							\
}
/*}}}*/

/*{{{  AddListN*/
/* 	Add non-empty 'list' to the head of the specified queue
*/
#define AddListN( q, list )	{					\
	QUEUE_CHAIN *temp;						\
	temp = list;							\
	while (QNext(temp)!= NULL) temp = QNext(temp);			\
	((QUEUE_CHAIN *)temp)->Next = q;				\
	q = list;							\
}

#define ListLength( q, len ){						\
	QUEUE_CHAIN *temp;						\
	temp = q;len = 0;						\
	while ( temp != NULL ) {					\
		temp = QNext(temp); len++;				\
	}								\
}
/*}}}*/

