#define MAX_NODES		100
#define MAX_TASKS		2

#define NO_ACK			1
#define PACKETIZE		1
#define PACKET_SIZE		(16*1024)



#define MIN_VIRTUAL_CHANNEL	0x80000040
#define BYTES_PER_VIRTUAL_CHANNEL	8
#define NOT_PROCESS		0x80000000

/*	Virtual Channel structure :
	0:	Loader
	1.. MAX_TASKS : Used by Task i to talk to the server
	MAX_TASKS*(NUM_NODES-1) channels are used by each task to talk to
		the tasks on other nodes
*/
#define MIN_PVM_VIRTUAL_CHANNEL		((unsigned long)MIN_VIRTUAL_CHANNEL + ( 1+ MAX_TASKS )*BYTES_PER_VIRTUAL_CHANNEL)

/* Internal channels to communicate between tasks on the same node are
   allocated just below the loader
   #define MIN_PVM_INTERNAL_CHANNEL	0x807fd000
   MAX_TASKS channels are allocated for each task to talk to the other tasks
   on this node
*/
#define BYTES_PER_INTERNAL_CHANNEL 4

#define PACKET_LENGTH			32


#define EXTRA_BYTES_1		0
#define EXTRA_BYTES_2		0
#define EXTRA_BYTES		(EXTRA_BYTES_1 + EXTRA_BYTES_2)

#define DEFAULT_ISERVER_STACKSIZE	4096
#define DEFAULT_PVM_STACKSIZE	2048
#define DEFAULT_STACKSIZE	1024

#define SUCCESS			0

#define ACK_MESSAGE		0
#define ASYNC_MESSAGE		1
#define SYNC_MESSAGE		2
#define EXIT_MESSAGE		3
#define STOP_MESSAGE		4


#define MAX_TCL_COMMAND		512
#define MAX_RECEIVE_REQUESTS	100
#define MAX_MESSAGES_QUEUED	100
#define MAX_PACKET_BUFFERS	20

#define FIXUP_LEN( MessageLen )	MessageLen
/*((MessageLen <= PACKET_LENGTH)?MessageLen:\
		(((MessageLen + PACKET_LENGTH - 1)/PACKET_LENGTH)*PACKET_LENGTH+1) )
*/


/*      BOOLEAN type
*/
typedef enum { True = 1, False = 0 } Boolean;

typedef short   INT16;

/*      Process descriptor on a transputer
*/
typedef long WDESC_TYPE;

/*      Priorities in a transputer
*/
typedef enum { High=0, Low=1 }PRIORITY_TYPE;




/*      Queue : Should always be inserted as the first element
        of the structure that is being queued.
*/
typedef struct queue_chain{
        struct queue_chain *Next;
}QUEUE_CHAIN;

typedef struct receive_info {
	long		tag;
	long 		MessageLength;
	long		SourceNode;
	long		SourceTask;
	long		NotUsed[4];
}RECEIVE_INFO;

/*      Queue of posted receive requests
        NOTE : qlink should be the first element of the structure
*/
typedef struct receive_request {
        QUEUE_CHAIN     qlink;          /* Next receive request on this vchan */
        int             BufferSize;     /* Expected message size */
        unsigned char   *BufferPtr;     /* Receive buffer */
        WDESC_TYPE      ReceiverProcess;/* Process id of receiver */
	int		SourceNode;
	int		SourceTask;
	int		tag;
	RECEIVE_INFO	*info;
}RECEIVE_REQUEST;

typedef struct ack_wait {
	QUEUE_CHAIN	qlink;
	WDESC_TYPE	WaitingProcess;
	int		MessageId;
}ACK_WAIT;

typedef struct message_header {
	int		MessageLength;
	int		tag;
	int		flag;	/* SYNC/ASYNC/ACK */
#ifndef NO_ACK
	ACK_WAIT	*AckWait;
#endif
}MESSAGE_HEADER;

typedef struct packet_buffer {
	QUEUE_CHAIN	qlink;
	Boolean AllocFlag;
#if 0
	unsigned char pad1[EXTRA_BYTES_1];
#endif
	unsigned char Data[PACKET_SIZE];
#if 0
	unsigned char pad2[EXTRA_BYTES_2];
#endif
}PACKET_BUFFER;

typedef struct receive_buffer {
	QUEUE_CHAIN	qlink;
	Boolean AllocFlag;
#if 0
	unsigned char pad1[EXTRA_BYTES_1];
#endif
	unsigned char Data[1];
#if 0
	unsigned char pad2[EXTRA_BYTES_2];
#endif
}RECEIVE_BUFFER;

typedef struct message {
	QUEUE_CHAIN	qlink;
	MESSAGE_HEADER	Header;
	RECEIVE_BUFFER	*Buffer;
	int		SourceTask;
	int		SourceNode;
	Boolean		AllocFlag;
}MESSAGE;
	
typedef struct {
	long	sCount;
	unsigned long sFront;
	unsigned long sBack;
}_Semaphore;


typedef struct ack_struct {
	QUEUE_CHAIN 	qlink;
	int		NodeId;
	int		TaskId;
	MESSAGE_HEADER	header;
}ACK_STRUCT;

#if 0
typedef struct pvmd_parameters {
	int		NumNodes;
	char		*TclSocket;
	int		*NumTasksLoaded;
	char		**Header;
	char		**AckHeader;
	char		**ServHeader;
	char		**ServAckHeader;
}PVMD_PARAMETERS;
#else
typedef struct pvmd_parameters {
	int		NumNodes;
	int		TclSocket;
	char		*NumTasksLoaded;
	char		**Header;
	char		**AckHeader;
	char		**ServHeader;
	char		**ServAckHeader;
}PVMD_PARAMETERS;
#endif


typedef struct stdio_struct {
	QUEUE_CHAIN 	qlink;
	int tid;
	int dst;
	int cod;
	int len;
	char *buf;
}StdIOStruct;
	
