
static char rcsid[] = 
	"$Id: trace.c,v 2.310 1994/07/18 18:18:15 kohl Exp $";

#include "xpvm.h"

read_trace_event( tsec, tusec, tid, eid )
int *tsec;
int *tusec;
int *tid;
int *eid;
{
	char tmp[4096];
	char c;

	int found;
	int done;
	int flag;
	int ret;
	int i;

	if ( TRACE_IN == NULL )
		return( FALSE );

	done = 0;

	ret = FALSE;

	while ( !done )
	{
		flag = fscanf( TRACE_IN, "%c", &c );
	
		if ( flag == 1 )
		{
			/* printf( "<%c>", c ); */

			switch ( c )
			{
				/* Command Packet */
	
				case '%':
				{
					if ( (flag = find_event_end()) != 1 )
						done--;

					break;
				}
	
				/* Record Descriptor Packet */
	
				case '#':
				{
					if ( (flag = find_event_end()) != 1 )
						done--;

					break;
				}
	
				/* Stream Attribute Packet */
	
				case '/':
				{
					flag = fscanf( TRACE_IN, "%c", &c );

					if ( flag == 1 )
					{
						if ( c != '*' )
						{
							printf( "\nError Reading Trace File\n" );
							printf( "\t- '*' expected (/* */)\n\n" );
						}

						if ( (flag = find_event_end()) != 1 )
							done--;
					}

					else
						done--;

					break;
				}
	
				/* White Space */

				case ' ':
				case '\t':
				case '\n':
				{
					break;
				}

				/* Record Data Packet */
	
				case '"':
				{
					i = 0;

					while ( (flag = fscanf( TRACE_IN, "%c", &c )) == 1
						&& c != '"' )
					{
						tmp[i++] = c;
					}

					if ( flag == 1 )
					{
						tmp[i] = '\0';

						/* printf( "{%s}", tmp ); */

						found = 0;

						for ( i=0 ; i < TRACE_MAX && !found ; i++ )
						{
							if ( !strcmp( tmp, TEV_TRACE_NAMES[i] ) )
								found++;
						}

						if ( found )
						{
							*eid = i - 1;

							/* printf( "%d ", *eid ); */

							while ( (flag=fscanf( TRACE_IN, "%c", &c ))
								== 1 && c != '{' );

							/* } to match above */

							if ( flag == 1 )
							{
								flag = fscanf( TRACE_IN, "%d, %d, %d",
									tsec, tusec, tid );

								if ( flag == 3 )
								{
									ret = TRUE;

									done++;
								}

								else
									done--;
							}

							else
								done--;
						}

						else
						{
							printf(
								"\nError: Event \"%s\" Not Found\n\n",
								tmp );

							if ( (flag = find_event_end()) != 1 )
								done--;

							else
								done++;
						}
					}

					else
						done--;

					break;
				}
			}
		}

		else
		{
			if ( TRACE_FILE_STATUS == TRACE_FILE_PLAYBACK )
				done++;

			else
				done--;
		}
	}

	if ( done < 0 )
		printf( "\nError Reading Trace File\n\n" );

	return( ret );
}

store_trace_event( tsec, tusec, tid, eid )
int tsec;
int tusec;
int tid;
int eid;
{
	char where[4096];
	char name[4096];

	int *tids;

	int msgtag;
	int nbytes;
	int count;
	int flags;
	int addr;
	int host;
	int type;
	int ptid;
	int dtid;
	int stid;
	int len;
	int buf;
	int num;
	int cc;
	int i;

	/* Output Trace Event Descriptor (if necessary) */

	if ( !( TEV_TRACE[ eid ] ) )
	{
		fprintf( TRACE_OUT, "\n#%d: %s\n",
			eid + 1, TEV_TRACE_DESCRIPTORS[ eid ] );

		(TEV_TRACE[ eid ])++;
	}

	/* Output Trace Event Header */

	fprintf( TRACE_OUT, "\"%s\" { %d, %d, %d",
		TEV_TRACE_NAMES[ eid ], tsec, tusec, tid );  /* } */

	/* Convert Trace to Output File */

	switch ( eid )
	{
		case TEV_NEWTASK:
		{
			pvm_upkint( &ptid, 1, 1 );
			pvm_upkint( &flags, 1, 1 );
			pvm_upkstr( name );

			if ( !strcmp( name, "" ) )
				strcpy( name, "-" );

			host = pvm_tidtohost( tid );

			fprintf( TRACE_OUT, ", %d, %d, %d",
				ptid, host, flags );

			dump_trace_str( name, TRUE );

			break;
		}

		case TEV_ADDHOSTS0:
		{
			pvm_upkint( &num, 1, 1 );

			fprintf( TRACE_OUT, ", %d, [%d][100] { ", num, num );

			for ( i=0 ; i < num ; i++ )
			{
				pvm_upkstr( name );

				if ( i != num - 1 )
					fprintf( TRACE_OUT, "\"%s\", ", name );
				
				else
					fprintf( TRACE_OUT, "\"%s\" }", name );
			}

			break;
		}

		case TEV_ADDHOSTS1:
		{
			pvm_upkint( &cc, 1, 1 );

			fprintf( TRACE_OUT, ", %d", cc );

			break;
		}

		case TEV_SPAWN0:
		{
			pvm_upkstr( name );
			pvm_upkint( &flags, 1, 1 );
			pvm_upkstr( where );
			pvm_upkint( &count, 1, 1 );

			dump_trace_str( name, TRUE );

			fprintf( TRACE_OUT, ", %d", flags );

			dump_trace_str( where, TRUE );

			fprintf( TRACE_OUT, ", %d", count );

			break;
		}

		case TEV_SPAWN1:
		{
			pvm_upkint( &num, 1, 1 );

			fprintf( TRACE_OUT, ", %d, [%d] {", num, num );

			if ( num > 0 )
			{
				tids = (int *) malloc( (unsigned) num * sizeof(int) );
				memcheck( tids, "Spawn TIDs" );

				pvm_upkint( tids, num, 1 );

				for ( i=0 ; i < num ; i++ )
				{
					if ( i )
						fprintf( TRACE_OUT, ", %d", tids[i] );
					
					else
						fprintf( TRACE_OUT, " %d", tids[i] );
				}

				free( tids );
			}

			fprintf( TRACE_OUT, " }" );

			break;
		}

		case TEV_SEND0:
		{
			pvm_upkint( &dtid, 1, 1 );
			pvm_upkint( &msgtag, 1, 1 );

			fprintf( TRACE_OUT, ", %d, %d", dtid, msgtag );

			break;
		}

		case TEV_MCAST0:
		{
			pvm_upkint( &num, 1, 1 );
			pvm_upkint( &msgtag, 1, 1 );

			fprintf( TRACE_OUT, ", %d, %d, [%d] {", num, msgtag, num );

			if ( num > 0 )
			{
				tids = (int *) malloc( (unsigned) num * sizeof(int) );
				memcheck( tids, "Spawn TIDs" );

				pvm_upkint( tids, num, 1 );

				for ( i=0 ; i < num ; i++ )
				{
					if ( i )
						fprintf( TRACE_OUT, ", %d", tids[i] );
					
					else
						fprintf( TRACE_OUT, " %d", tids[i] );
				}

				free( tids );
			}

			fprintf( TRACE_OUT, " }" );

			break;
		}

		case TEV_PSEND0:
		{
			pvm_upkint( &dtid, 1, 1 );
			pvm_upkint( &msgtag, 1, 1 );
			pvm_upkint( &addr, 1, 1 );
			pvm_upkint( &len, 1, 1 );
			pvm_upkint( &type, 1, 1 );

			fprintf( TRACE_OUT, ", %d, %d, %d, %d, %d",
				dtid, msgtag, addr, len, type );

			break;
		}

		case TEV_RECV1:
		case TEV_TRECV1:
		{
			pvm_upkint( &buf, 1, 1 );
			pvm_upkint( &nbytes, 1, 1 );

			if ( nbytes >= 0 )
			{
				pvm_upkint( &msgtag, 1, 1 );
				pvm_upkint( &stid, 1, 1 );
			}

			else
				msgtag = stid = -1;

			fprintf( TRACE_OUT, ", %d, %d, %d, %d",
				buf, nbytes, msgtag, stid );

			break;
		}

		case TEV_PRECV1:
		/* temporary handler til precv1 has proper contents */
		{
			pvm_upkint( &buf, 1, 1 );

			fprintf( TRACE_OUT, ", %d, 0, 0, 0", buf );

			break;
		}


		case TEV_NRECV1:
		{
			pvm_upkint( &buf, 1, 1 );

			if ( buf > 0 )
			{
				pvm_upkint( &nbytes, 1, 1 );

				if ( nbytes >= 0 )
				{
					pvm_upkint( &msgtag, 1, 1 );
					pvm_upkint( &stid, 1, 1 );
				}

				else
					msgtag = stid = -1;
			}

			else
				nbytes = msgtag = stid = -1;

			fprintf( TRACE_OUT, ", %d, %d, %d, %d",
				buf, nbytes, msgtag, stid );

			break;
		}

		default:
			dump_pvm_event_fmt( eid );
	}

	/* Finish Off Trace Event ({) */

	fprintf( TRACE_OUT, " };;\n" );
}

dump_pvm_event_fmt( eid )
int eid;
{
	char buf[4096];

	char *fmt;

	int keeplist[10];

	int vecflag;
	int veclen;
	int index;
	int nkeep;
	int keep;
	int ival;

	fmt = tev_formats[ eid ].fmt;

	if ( fmt != NULL )
	{
		nkeep = 0;

		while ( *fmt )
		{
			if ( *fmt == '%' )
			{
				fmt++;

				if ( *fmt == '*' )
				{
					keep = TRUE;

					fmt++;
				}

				else
					keep = FALSE;

				if ( *fmt == '$' )
				{
					fmt++;

					index = *fmt - '0';

					fmt++;

					veclen = ( index >= 0 && index < nkeep )
						? keeplist[index] : 0;

					vecflag = TRUE;
				}

				else
				{
					veclen = 1;

					vecflag = FALSE;
				}

				while ( veclen > 0 )
				{
					veclen--;

					switch ( *fmt )
					{
						case 'R':
						case 'd':
						case 'r':
						case 'x':
						{
							pvm_upkint( &ival, 1, 1 );
		
							fprintf( TRACE_OUT, ", %d", ival );
		
							break;
						}
		
						case 'S':
						{
							pvm_upkstr( buf );
		
							dump_trace_str( buf, TRUE );
		
							break;
						}
		
						default:
							break;
					}
				}
	
				if ( *fmt )
					fmt++;

				if ( keep )
				{
					keeplist[nkeep] = ival < 0 ? 0 : ival;

					nkeep++;
				}
			}
			
			else
				fmt++;
		}
	}
}

store_output_event( tid, str, len )
int tid;
char *str;
int len;
{
	char tmp[4096];

	int index;
	int num;
	int i;

	index = 0;

	num = 0;

	for ( i=0 ; i < len ; i++ )
	{
		if ( str[i] != '\n' )
			tmp[index++] = str[i];

		else
		{
			tmp[index] = '\0';

			write_output_event( tid, tmp );

			index = 0;

			num++;
		}
	}

	if ( index )
	{
		tmp[index] = '\0';

		write_output_event( tid, tmp );

		num++;
	}

	return( num );
}

write_output_event( tid, str )
int tid;
char *str;
{
	if ( !( TEV_TRACE[ TRACE_OUTPUT ] ) )
	{
		fprintf( TRACE_OUT, "\n#%d: %s\n",
			TRACE_OUTPUT + 1, TEV_TRACE_DESCRIPTORS[ TRACE_OUTPUT ] );

		(TEV_TRACE[ TRACE_OUTPUT ])++;
	}

	fprintf( TRACE_OUT, "\"%s\" { %d, %d, %d",
		TEV_TRACE_NAMES[ TRACE_OUTPUT ],
		TRACE_TIME.tv_sec, TRACE_TIME.tv_usec, tid );

	dump_trace_str( str, TRUE );

	fprintf( TRACE_OUT, " };;\n" );
}

save_host_status_events()
{
	HOST H;

	int sync;
	int num;

	if ( TRACE_OUT == NULL )
		return( 0 );

	if ( !( TEV_TRACE[ TRACE_HOST_ADD ] ) )
	{
		fprintf( TRACE_OUT, "\n#%d: %s\n",
			TRACE_HOST_ADD + 1,
			TEV_TRACE_DESCRIPTORS[ TRACE_HOST_ADD ] );

		(TEV_TRACE[ TRACE_HOST_ADD ])++;
	}

	H = HOST_LIST;

	sync = 0;

	num = 0;

	while ( H != NULL )
	{
		if ( H->in_pvm == IN_PVM )
		{
			write_host_add_event( H );

			if ( H->delta.tv_sec != 0 || H->delta.tv_usec != 0 )
				sync++;

			num++;
		}

		H = H->next;
	}

	if ( sync )
	{
		H = HOST_LIST;
	
		while ( H != NULL )
		{
			if ( H->in_pvm == IN_PVM &&
				( H->delta.tv_sec != 0 || H->delta.tv_usec != 0 ) )
			{
				write_host_sync_event( H );
	
				num++;
			}
	
			H = H->next;
		}
	}

	if ( num )
		fflush( TRACE_OUT );

	return( num );
}

write_host_add_event( H )
HOST H;
{
	if ( !( TEV_TRACE[ TRACE_HOST_ADD ] ) )
	{
		fprintf( TRACE_OUT, "\n#%d: %s\n",
			TRACE_HOST_ADD + 1,
			TEV_TRACE_DESCRIPTORS[ TRACE_HOST_ADD ] );

		(TEV_TRACE[ TRACE_HOST_ADD ])++;
	}

	fprintf( TRACE_OUT, "\"%s\" { %d, %d, %d",
		TEV_TRACE_NAMES[ TRACE_HOST_ADD ],
		TRACE_TIME.tv_sec, TRACE_TIME.tv_usec, H->pvmd_tid );

	dump_trace_str( H->refname, TRUE );

	dump_trace_str( H->alias, TRUE );

	dump_trace_str( H->arch, TRUE );

	fprintf( TRACE_OUT, ", %d };;\n", H->speed );
}

store_host_del_event( name, tid )
char *name;
int tid;
{
	if ( !( TEV_TRACE[ TRACE_HOST_DEL ] ) )
	{
		fprintf( TRACE_OUT, "\n#%d: %s\n",
			TRACE_HOST_DEL + 1,
			TEV_TRACE_DESCRIPTORS[ TRACE_HOST_DEL ] );

		(TEV_TRACE[ TRACE_HOST_DEL ])++;
	}

	fprintf( TRACE_OUT, "\"%s\" { %d, %d, %d",
		TEV_TRACE_NAMES[ TRACE_HOST_DEL ],
		TRACE_TIME.tv_sec, TRACE_TIME.tv_usec, tid );

	dump_trace_str( name, TRUE );

	fprintf( TRACE_OUT, " };;\n" );
}

write_host_sync_event( H )
HOST H;
{
	if ( !( TEV_TRACE[ TRACE_HOST_SYNC ] ) )
	{
		fprintf( TRACE_OUT, "\n#%d: %s\n",
			TRACE_HOST_SYNC + 1,
			TEV_TRACE_DESCRIPTORS[ TRACE_HOST_SYNC ] );

		(TEV_TRACE[ TRACE_HOST_SYNC ])++;
	}
	
	fprintf( TRACE_OUT, "\"%s\" { %d, %d, %d, %d, %d };;\n",
		TEV_TRACE_NAMES[ TRACE_HOST_SYNC ],
		TRACE_TIME.tv_sec, TRACE_TIME.tv_usec, H->pvmd_tid,
		H->delta.tv_sec, H->delta.tv_usec );
}

dump_trace_header()
{
	char *ds;

	if ( TRACE_OUT == NULL )
		return;

	fprintf( TRACE_OUT, "/*\n" );

	ds = date_str();

	fprintf( TRACE_OUT, " * \"Creation Date\" \"%s\"\n", ds );

	free( ds );

	fprintf( TRACE_OUT, " * \"Machine\" \"XPVM %s (PVM %s)\"\n",
		XPVM_VERSION, pvm_version() );

	fprintf( TRACE_OUT, " */ ;;\n\n" );
}

dump_trace_str( str, comma_flag )
char *str;
int comma_flag;
{
	if ( comma_flag )
	{
		fprintf( TRACE_OUT, ", [%d] { \"%s\" }",
			strlen( str ) + 1, str );
	}
	
	else
		fprintf( TRACE_OUT, "[%d] { \"%s\" }", strlen( str ) + 1, str );
}

dump_sddf_headers()
{
	int eid;

	printf( "\n" );
	printf( "XPVM Trace Format Descriptors - SDDF:\n" );
	printf( "=====================================\n" );

	for ( eid=0 ; eid < TRACE_MAX ; eid++ )
	{
		if ( strcmp( TEV_TRACE_DESCRIPTORS[ eid ], "" ) )
		{
			printf( "\n#%d: %s\n", eid + 1,
				TEV_TRACE_DESCRIPTORS[ eid ] );
		}
	}

	exit( 0 );
}

read_trace_str( str, size, end_flag )
char *str;
int size;
int end_flag;
{
	char c;

	int i;

	while ( fscanf( TRACE_IN, "%c", &c ) == 1 && c != '"' );

	if ( c == '"' )
	{
		i = 0;

		while ( fscanf( TRACE_IN, "%c", &c ) == 1 && c != '"'
			&& i < size - 1 )
		{
			str[i++] = c;
		}

		str[i] = '\0';

		if ( c != '"' )
			while ( fscanf( TRACE_IN, "%c", &c ) == 1 && c != '"' );

		if ( end_flag && c == '"' )
		{
			/* Matching { */

			while ( fscanf( TRACE_IN, "%c", &c ) == 1 && c != '}' );
		}
	}

	else
		*str = '\0';
}

check_overwrite_exists( itp )
Tcl_Interp *itp;
{
	FILE *fp;

	char tmp[1024];

	if ( TRACE_OVERWRITE_FLAG )
		return( TRUE );

	fp = fopen( TRACE_FILE, "r" );

	if ( fp != NULL )
	{
		fclose( fp );

		sprintf( tmp, "prompt_trace_overwrite %s", TRACE_FILE );

		Tcl_Eval( itp, tmp );

		return( FALSE );
	}

	else
		return( TRUE );
}

find_event_end()
{
	char last;
	char c;

	int flag;

	last = '\0';

	while ( (flag = fscanf( TRACE_IN, "%c", &c )) == 1 )
	{
		/* printf( "[%c]", c ); */

		if ( c == ';' && last == ';' )
			return( 1 );

		last = c;
	}

	return( flag );
}

