
static char rcsid[] = 
	"$Id: util.c,v 2.310 1994/07/18 18:18:17 kohl Exp $";

#include "xpvm.h"

MSG create_msg()
{
	MSG tmp;

	tmp = (MSG) malloc( sizeof( struct msg_struct ) );
	memcheck( tmp, "Message Structure" );

	tmp->src_tid = -1;
	tmp->dst_tid = -1;
	tmp->msgtag = -1;
	tmp->nbytes = -1;

	tmp->sendtime.tv_sec = -1;
	tmp->sendtime.tv_usec = -1;

	tmp->recvtime.tv_sec = -1;
	tmp->recvtime.tv_usec = -1;

	tmp->next = (MSG) NULL;

	return( tmp );
}

free_msg( ptr )
MSG *ptr;
{
	MSG M;

	M = *ptr;

	M->src_tid = -1;
	M->dst_tid = -1;
	M->msgtag = -1;
	M->nbytes = -1;

	M->sendtime.tv_sec = -1;
	M->sendtime.tv_usec = -1;

	M->recvtime.tv_sec = -1;
	M->recvtime.tv_usec = -1;

	M->next = (MSG) NULL;

	free( M );

	*ptr = (MSG) NULL;
}

TASK create_task()
{
	TASK tmp;

	tmp = (TASK) malloc( sizeof( struct task_struct ) );
	memcheck( tmp, "Task Structure" );

	tmp->name = (char *) NULL;

	tmp->tid = -1;
	tmp->ptid = -1;
	tmp->pvmd_tid = -1;
	tmp->flags = -1;
	tmp->status = -1;

	tmp->next = (TASK) NULL;

	return( tmp );
}

free_task( ptr )
TASK *ptr;
{
	TASK T;

	T = *ptr;

	if ( T->name != NULL )
		free( T->name );
	
	T->tid = -1;
	T->ptid = -1;
	T->pvmd_tid = -1;
	T->flags = -1;

	T->next = (TASK) NULL;

	free( T );

	*ptr = (TASK) NULL;
}

TASK get_task_tid( tid )
int tid;
{
	TASK T;

	T = TASK_LIST;

	while ( T != NULL )
	{
		if ( T->tid == tid )
			return( T );

		T = T->next;
	}

	return( (TASK) NULL );
}

tasks_alive()
{
	TASK T;

	T = TASK_LIST;

	while ( T != NULL )
	{
		if ( T->status == TASK_ALIVE )
			return( TRUE );

		T = T->next;
	}

	return( FALSE );
}

HOST create_host()
{
	HOST tmp;

	tmp = (HOST) malloc( sizeof( struct host_struct ) );
	memcheck( tmp, "Host Structure" );

	tmp->name = (char *) NULL;

	tmp->alias = (char *) NULL;

	tmp->refname = (char *) NULL;

	tmp->arch = (char *) NULL;

	tmp->delta.tv_usec = 0;
	tmp->delta.tv_sec = 0;

	tmp->pvmd_tid = -1;
	tmp->in_pvm = -1;
	tmp->index = -1;
	tmp->speed = -1;

	tmp->next = (HOST) NULL;

	return( tmp );
}

HOST get_host_index( index )
int index;
{
	HOST H;

	HOST tmp;

	tmp = (HOST) NULL;

	H = HOST_LIST;

	while ( H != NULL && tmp == NULL )
	{
		if ( H->index == index )
			tmp = H;

		else
			H = H->next;
	}

	return( tmp );
}

ARCH create_arch()
{
	ARCH tmp;

	tmp = (ARCH) malloc( sizeof( struct arch_struct ) );
	memcheck( tmp, "Architecture Structure" );

	tmp->code = (char *) NULL;

	tmp->name = (char *) NULL;

	tmp->height = -1;
	tmp->width = -1;

	tmp->next = (ARCH) NULL;

	return( tmp );
}

ARCH get_arch_code( name )
char *name;
{
	ARCH A;

	ARCH tmp;

	if ( name == NULL )
		return( (ARCH) NULL );

	tmp = (ARCH) NULL;

	A = ARCH_LIST;

	while ( A != NULL && tmp == NULL )
	{
		if ( !strcmp( name, A->code ) )
			tmp = A;

		else
			A = A->next;
	}

	if ( tmp == NULL )
	{
		printf( "\nWarning: Unknown or Missing Architecture %s\n\n",
			name );
	}

	return( tmp );
}

/*
ARCH get_arch_id( id )
int id;
{
	ARCH A;

	ARCH tmp;

	tmp = (ARCH) NULL;

	A = ARCH_LIST;

	while ( A != NULL && tmp == NULL )
	{
		if ( A->id == id )
			tmp = A;

		else
			A = A->next;
	}

	if ( tmp == NULL )
		printf( "\nWarning: Unknown or Missing Arch ID %d\n\n", id );

	return( tmp );
}
*/

char *get_arch_icon_str( A )
ARCH A;
{
	char str[1024];

	char *tmp;

	if ( A != NULL )
		sprintf( str, "%s %d %d", A->name, A->width, A->height );

	else
	{
		A = get_arch_code( "DUMMY" );

		sprintf( str, "%s %d %d", A->name, A->width, A->height );
	}

	tmp = copy_str( str );

	return( tmp );
}

TRACE_MASK_INDEX create_trace_mask_index()
{
	TRACE_MASK_INDEX tmp;

	tmp = (TRACE_MASK_INDEX) malloc(
		sizeof( struct trace_mask_index_struct ) );
	memcheck( tmp, "Trace Mask Index Structure" );

	tmp->name = (char *) NULL;

	tmp->index = -1;

	tmp->next = (TRACE_MASK_INDEX) NULL;

	return( tmp );
}

read_hostfile()
{
	HOST H;

	FILE *fp;

	char hostline[1024];
	char reftmp[1024];
	char msg[1024];
	char tmp[1024];

	char *tmp_alias;

	int in_pvm;
	int done;
	int i;

	/* Add local host to list */

	if ( NEW_PVM && HOST_NAME != NULL )
	{
		H = HOST_LIST = create_host();

		H->name = copy_str( HOST_NAME );

		H->alias = copy_str( HOST_ALIAS );

		H->refname = copy_str( H->name );

		H->in_pvm = TO_BE_ADDED;

		H->index = 0;

		NHOSTS = 1;
	}

	/* Get list of hosts */

	fp = fopen( HOSTFILE, "r" );

	if ( fp != NULL )
	{
		done = 0;

		do
		{
			i = 0;

			while ( fscanf( fp, "%c", &(hostline[i]) ) == 1
				&& hostline[i] != '\n' )
			{
				i++;
			}

			if ( hostline[i] != '\n' )
				done++;

			else
			{
				hostline[i] = '\0';

				if ( hostline[0] != '#' && hostline[0] != '*'
					&& strcmp( hostline, "" ) )
				{
					if ( hostline[0] == '&' )
					{
						sprintf( reftmp, "%s", hostline + 1 );
	
						in_pvm = NOT_IN_PVM;
					}
	
					else
					{
						sprintf( reftmp, "%s", hostline );
	
						in_pvm = TO_BE_ADDED;
					}
	
					sscanf( reftmp, "%s", tmp );

					tmp_alias = host_alias_str( tmp );
	
					if ( !LOCALHOST( tmp ) && !LOCALHOST( tmp_alias ) )
					{
						if ( HOST_LIST == NULL )
						{
							H = HOST_LIST = create_host();
	
							NHOSTS = 0;
						}
	
						else
						{
							H->next = create_host();
	
							H = H->next;
						}
	
						H->name = copy_str( tmp );
	
						H->alias = tmp_alias;
	
						H->refname = copy_str( reftmp );
	
						H->in_pvm = in_pvm;
	
						H->index = NHOSTS;
	
						NHOSTS++;
					}
	
					else
						free( tmp_alias );
				}
			}
		}
		while ( !done );

		fclose( fp );
	}

	else
	{
		sprintf( tmp, "Error Opening Hostfile \"%s\"", HOSTFILE );

		sprintf( msg, "setMsg { %s }", tmp );

		Tcl_Eval( interp, msg );

		fprintf( stderr, "\n%s\n\n", tmp );
	}
}

initialize_hosts()
{
	HOST H;

	ARCH A;

	HOST last;

	struct pvmhostinfo *HP;

	struct pvmhostinfo *hostp;

	char result[2048];
	char tmp[4096];

	char *hostalias;
	char *iconstr;

	int localck;
	int found;
	int narch;
	int nhost;
	int host;
	int cc;
	int i;

	strcpy( result, "" );

	/* Get Existing Configuration */

	cc = pvm_config( &nhost, &narch, &hostp );

	if ( cc < 0 )
	{
		pvm_perror( "Checking Existing Configuration" );

		error_exit();
	}

	/* Set up Host Delete Notify's */

	for ( i=0 ; i < nhost ; i++ )
		pvm_notify( PvmHostDelete, 100, 1, &(hostp[i].hi_tid) );

	/* Check Existing List Against Host File */

	for ( i=0 ; i < nhost ; i++ )
	{
		HP = &(hostp[i]);

		/* Update Host List */

		H = HOST_LIST;

		last = (HOST) NULL;

		found = 0;

		while ( H != NULL && !found )
		{
			hostalias = host_alias_str( HP->hi_name );

			if ( !strcmp( H->name, HP->hi_name )
				|| !strcmp( H->alias, HP->hi_name )
				|| !strcmp( H->name, hostalias )
				|| !strcmp( H->alias, hostalias ) )
			{
				found++;
			}

			else
			{
				last = H;

				H = H->next;
			}

			free( hostalias );
		}

		/* Host Already in Host List, Set Fields */

		if ( found )
		{
			free( H->name );

			free( H->alias );
			
			H->name = copy_str( HP->hi_name );

			H->alias = host_alias_str( H->name );

			H->arch = check_arch( H, HP->hi_arch );

			H->pvmd_tid = HP->hi_tid;

			H->speed = HP->hi_speed;

			H->in_pvm = IN_PVM;
		}

		/* Host Not Found, Add to Hosts List */

		else
		{
			if ( last != NULL )
				H = last->next = create_host();
			
			else
			{
				H = HOST_LIST = create_host();

				NHOSTS = 0;
			}

			H->name = copy_str( HP->hi_name );

			H->alias = host_alias_str( HP->hi_name );

			H->refname = copy_str( H->name );

			H->arch = check_arch( H, HP->hi_arch );

			H->pvmd_tid = HP->hi_tid;

			H->speed = HP->hi_speed;

			H->index = NHOSTS;

			H->in_pvm = IN_PVM;

			sprintf( tmp, "{NEW %s %d PERM} ",
				host_name( H, FALSE ), MAX_HID + NHOSTS );

			append_cmd_str( result, tmp, 2048,
				"checkHostResult", TRUE );

			NHOSTS++;
		}

		/* Get Icon Info */

		A = get_arch_code( H->arch );

		iconstr = get_arch_icon_str( A );

		sprintf( tmp, "{ICON %d %s} ", H->index, iconstr );

		append_cmd_str( result, tmp, 2048,
			"checkHostResult", TRUE );

		free( iconstr );

		/* Turn Host On */

		sprintf( tmp, "{ON %d} ", H->index );

		append_cmd_str( result, tmp, 2048,
			"checkHostResult", TRUE );
		
		/* Set Host Delay */

		if ( !LOCALHOST( H->name ) )
		{
			pvm_hostsync( H->pvmd_tid,
				(struct timeval *) NULL, &(H->delta) );
		}
	}

	sprintf( tmp, "checkHostResult { %s }", result );

	Tcl_Eval( interp, tmp );

	/* Add in Any Missed Hosts from Host File */

	localck = 0;

	H = HOST_LIST;

	while ( H != NULL )
	{
		if ( H->in_pvm == TO_BE_ADDED )
		{
			if ( !LOCALHOST( H->name ) )
			{
				H->in_pvm = NOT_IN_PVM;

				do_add_host( H );
			}

			else
			{
				H->in_pvm = IN_PVM;

				localck++;
			}
		}

		H = H->next;
	}

	/* Fix LOCALHOST / HOST_NAME - Network Name in use... */
	if ( !localck )
	{
		host = pvm_tidtohost( MYTID );

		H = HOST_LIST;

		while ( H != NULL && !localck )
		{
			if ( H->pvmd_tid == host )
			{
				if ( HOST_NAME != NULL )
					free( HOST_NAME );

				if ( HOST_ALIAS != NULL )
					free( HOST_ALIAS );

				HOST_NAME = copy_str( H->name );

				HOST_ALIAS = copy_str( H->alias );

				/* Verify Host Delay Cleared */

				H->delta.tv_sec = H->delta.tv_usec = 0;

				localck++;
			}

			H = H->next;
		}
	}
}

update_hosts( tids, num )
int *tids;
int num;
{
	HOST H;

	ARCH A;

	HOST last;

	struct pvmhostinfo *HP;

	struct pvmhostinfo *hostp;

	char result[2048];
	char tmp[4096];

	char *iconstr;

	int found;
	int narch;
	int nhost;
	int i, j;
	int cc;

	cc = pvm_config( &nhost, &narch, &hostp );

	if ( cc < 0 )
	{
		pvm_perror( "Checking Existing Configuration" );

		error_exit();
	}

	strcpy( result, "" );

	/* Look For New Hosts */

	for ( i=0 ; i < num ; i++ )
	{
		/* Find Hostinfo Struct */

		found = 0;

		for ( j=0 ; j < nhost && !found ; j++ )
		{
			HP = &(hostp[j]);

			if ( HP->hi_tid == tids[i] )
				found++;
		}

		if ( !found )
		{
			printf( "\nError: Missing Host Info tid=%x\n\n", tids[i] );

			continue;
		}

		/* Update Host List */

		H = HOST_LIST;

		last = (HOST) NULL;

		found = 0;

		while ( H != NULL && !found )
		{
			/* if ( H->pvmd_tid == HP->hi_tid ) */
			if ( !strcmp( H->name, HP->hi_name )
				|| !strcmp( H->alias, HP->hi_name ) )
			{
				found++;
			}

			else
			{
				last = H;

				H = H->next;
			}
		}

		/* Host Not Found, Add to Hosts List */

		if ( !found )
		{
			if ( last != NULL )
				H = last->next = create_host();
			
			else
			{
				H = HOST_LIST = create_host();

				NHOSTS = 0;
			}

			H->name = copy_str( HP->hi_name );

			H->alias = host_alias_str( HP->hi_name );

			H->refname = copy_str( H->name );

			H->arch = check_arch( H, HP->hi_arch );

			H->pvmd_tid = HP->hi_tid;

			H->speed = HP->hi_speed;

			H->index = NHOSTS;

			H->in_pvm = IN_PVM;

			sprintf( tmp, "{NEW %s %d PERM} ",
				host_name( H, FALSE ), MAX_HID + NHOSTS );

			append_cmd_str( result, tmp, 2048,
				"checkHostResult", TRUE );

			NHOSTS++;
		}

		/* Host Already in Host File, Just Update Fields */

		else
		{
			free( H->name );
			
			free( H->alias );

			H->name = copy_str( HP->hi_name );

			H->alias = host_alias_str( H->name );

			H->arch = check_arch( H, HP->hi_arch );

			H->pvmd_tid = HP->hi_tid;

			H->speed = HP->hi_speed;

			H->in_pvm = IN_PVM;
		}

		/* Get Icon Info */

		A = get_arch_code( H->arch );

		iconstr = get_arch_icon_str( A );

		sprintf( tmp, "{ICON %d %s} ", H->index, iconstr );

		append_cmd_str( result, tmp, 2048,
			"checkHostResult", TRUE );

		free( iconstr );

		/* Turn Host On */

		if ( TRACE_FILE_STATUS != TRACE_FILE_PLAYBACK )
		{
			sprintf( tmp, "{ON %d} ", H->index );

			append_cmd_str( result, tmp, 2048,
				"checkHostResult", TRUE );
		}

		/* Set Host Delay */

		if ( !LOCALHOST( H->name ) )
		{
			pvm_hostsync( H->pvmd_tid,
				(struct timeval *) NULL, &(H->delta) );
		}

		/* Save Host Add Event to Trace File */

		if ( TRACE_OUT != NULL )
		{
			update_trace_time( -1, -1 );

			write_host_add_event( H );

			TRACE_PENDING++;

			if ( H->delta.tv_sec != 0 || H->delta.tv_usec != 0 )
			{
				write_host_sync_event( H );

				TRACE_PENDING++;
			}

			fflush( TRACE_OUT );
		}
	}

	sprintf( tmp, "checkHostResult { %s }", result );

	Tcl_Eval( interp, tmp );
}

initialize_tasks()
{
	TASK T;

	struct pvmtaskinfo *TP;

	struct pvmtaskinfo *taskp;

	int ntask;
	int cc;
	int i;

	if ( NEW_PVM || TRUE )
		return;

	cc = pvm_tasks( 0, &ntask, &taskp );

	if ( cc < 0 )
	{
		pvm_perror( "Checking Initial Task Info" );

		error_exit();
	}

	for ( i=0 ; i < ntask ; i++ )
	{
		TP = &(taskp[i]);

		if ( TASK_LIST == NULL )
			T = TASK_LIST = create_task();
		
		else
		{
			T->next = create_task();

			T = T->next;
		}

		T->tid = TP->ti_tid;

		T->name = copy_str( TP->ti_a_out );

		T->ptid = TP->ti_ptid;

		T->pvmd_tid = TP->ti_host;

		T->flags = TP->ti_flag;

		T->status = TASK_ALIVE;

		add_task( T, 0, 0, "...On-Going Task" );
	}
}

set_trace_controls()
{
	char *str;

	switch ( TRACE_STATUS )
	{
		case TRACE_REWIND:	str = "trace_controls rewind";	break;

		case TRACE_STOP:	str = "trace_controls stop";	break;

		case TRACE_FWD:		str = "trace_controls fwd";		break;

		case TRACE_FWDSTEP:	str = "trace_controls fwdstep";	break;
	}

	Tcl_Eval( interp, str );
}

signal_by_name( name )
char *name;
{
	int i;

	for ( i=0 ; i < NUM_SIGNALS ; i++ )
	{
		if ( compare( SIGNAL_STRS[i], name ) )
			return( i + 1 );
	}

	return( -1 );
}

char *assemble_menu_list( nentries, labels, types, indices,
	states, icons )
int nentries;
char **labels;
char **types;
int *indices;
char **states;
char **icons;
{
	char **list_argv;

	char *result;

	char state[255];
	char icon[255];
	char tmp[1024];

	int i;

	list_argv = (char **) malloc( (unsigned) nentries
		* sizeof(char *) );
	memcheck( list_argv, "Menu List Argv" );

	for ( i=0 ; i < nentries ; i++ )
	{
		/* Grab State String (if any) */

		if ( states != NULL )
			sprintf( state, " %s", states[i] );
		
		else
			sprintf( state, "" );

		/* Grab Icon String (if any) */

		if ( icons != NULL )
			sprintf( icon, " %s", icons[i] );
		
		else
			sprintf( icon, "" );

		/* Assemble List String */

		sprintf( tmp, "\"%s\" %s %d%s%s",
			labels[i], types[i], indices[i], state, icon );

		list_argv[i] = copy_str( tmp );
	}

	result = Tcl_Merge( nentries, list_argv );

	return( result );
}

char *date_str()
{
	char tmp[255];

	char *result;

	time_t t;

	time( &t );

	sprintf( tmp, "%s", ctime( &t ) );

	result = tmp;

	while ( *result != '\n' && *result != '\0' )
		result++;
	
	*result = '\0';

	result = copy_str( tmp );

	return( result );
}

double distance( x1, y1, x2, y2 )
double x1, y1, x2, y2;
{
	return( sqrt(((x2 - x1) * (x2 - x1)) + ((y2 - y1) * (y2 - y1))) );
}

double angle_of( x1, y1, x2, y2 )
double x1, y1, x2, y2;
{
	double slope;
	double at;

	if ( x2 == x1 )
	{
		if ( y2 > y1 )
			return( PI / 2.0 );

		else
			return( -1.0 * PI / 2.0 );
	}

	slope = (y2 - y1) / (x2 - x1);

	at = atan(slope);

	if ( x1 > x2 )
	{
		if ( y2 > y1 )
			at += PI;

		else
			at -= PI;
	}

	return( at );
}

char *host_alias_str( name )
char *name;
{
	char *ptr;
	char *tmp;

	char c;

	if ( name == NULL )
		return( (char *) NULL );

	ptr = name;

	while ( *ptr != '\0' && *ptr != '.' )
		ptr++;

	if ( *ptr == '.' )
	{
		c = *ptr;

		*ptr = '\0';

		tmp = copy_str( name );

		*ptr = c;
	}

	else
		tmp = copy_str( name );
	
	return( tmp );
}

char *center_str( str, width )
char *str;
int width;
{
	char *tmp;

	int i, j;
	int len;

	tmp = (char *) malloc( (unsigned) (width + 1) * sizeof(char) );
	memcheck( tmp, "Center String" );

	len = strlen( str );

	if ( len > width )
	{
		strcpy( tmp, (char *) ( str + ( len - width ) ) );

		tmp[0] = '<';
	}

	else
	{
		j = ( width - len ) / 2;

		for ( i=0 ; i < j ; i++ )
			tmp[i] = ' ';

		strcpy( (char *) ( tmp + j ), str );

		for ( i=(j + len) ; i < width ; i++ )
			tmp[i] = ' ';

		tmp[width] = '\0';
	}

	return( tmp );
}

char *pad_num( num, max )
long num;
long max;
{
	char tmp[1024];

	char *ptr;
	char *str;

	int nd;
	int nz;
	int i;

	sprintf( tmp, "%ld", num );

	nd = strlen( tmp );

	nz = max - nd;

	str = (char *) malloc( (unsigned) (max + 1) * sizeof(char) );
	memcheck( str, "Numerical Pad String" );

	ptr = str;

	for ( i=0 ; i < nz ; i++ )
		*ptr++ = '0';

	sprintf( (char *) (str + nz), "%ld", num );

	return( str );
}

char *copy_str( str )
char *str;
{
	char *tmp;

	tmp = (char *) malloc( (unsigned) (strlen(str) + 1)
		* sizeof(char) );
	memcheck(tmp,"Copy String");

	strcpy( tmp, str );

	return( tmp );
}

append_cmd_str( result, str, length, err, cmd_flag )
char *result;
char *str;
int length;
char *err;
int cmd_flag;
{
	char *strcat();

	char *errstr;

	int elen;
	int rlen;
	int len;
	int rc;
	int i;

	rlen = strlen( result );

	len = rlen + strlen( str ) + 1;

	if ( len <= length )
		strcat( result, str );

	else
	{
		if ( cmd_flag == TRUE )
		{
			if ( err == NULL )
			{
				printf( "\nError in Append Cmd String: Null Cmd\n\n" );

				return( TCL_ERROR );
			}

			elen = len + strlen( err ) + 5;

			errstr = (char *) malloc( (unsigned) elen * sizeof(char) );
			memcheck( errstr, "Append Cmd String Err String" );

			sprintf( errstr, "%s { %s }", err, result );

			rc = Tcl_Eval( interp, errstr );

			if ( rc == TCL_ERROR )
				printf( "\nError Executing \"%s\"\n\n", errstr );

			free( errstr );

			if ( strlen( str ) <= length )
				strcpy( result, str );

			else
			{
				printf( "\nError Append Cmd String \"%s\" Too Long\n\n",
					str );

				strcpy( result, "" );
			}

			return( rc );
		}

		else
		{
			if ( err != NULL )
				printf( "\nError in Append Cmd String: %s\n\n", err );

			i = rlen;

			while ( i < length - 4 )
			{
				result[i] = str[ i - rlen ];

				i++;
			}

			sprintf( (char *) (result + (length - 4)), "..." );
		}
	}

	return( TCL_OK );
}

filecheck( fp, name )
FILE *fp;
char *name;
{
	char msg[1024];
	char tmp[1024];

	if ( fp == NULL )
	{
		sprintf( tmp, "Error Opening File \"%s\"", name );

		sprintf( msg, "setMsg { %s }", tmp );

		Tcl_Eval( interp, msg );

		fprintf( stderr, "\n%s\n\n", tmp );

		return( FALSE );
	}

	return( TRUE );
}

memcheck( ptr, name )
char *ptr;
char *name;
{
	if ( ptr == NULL )
	{
		fprintf( stderr, "\nError Allocating Memory for \"%s\"\n\n",
			name );

		error_exit();
	}
}

compare( x, y )
char *x, *y;
{
	while ( *x != '\0' && *y != '\0' )
	{
		if ( *x != *y )
			return( FALSE );

		x++; y++;
	}

	return( TRUE );
}

error_exit()
{
	int cc;

	if ( MYTID > 0 && GROUPS_ALIVE )
	{
		cc = pvm_getinst( "xpvm", MYTID );

		if ( cc >= 0 )
		{
			cc = pvm_lvgroup( "xpvm" );

			if ( cc < 0 )
			{
				printf( "\nError Leaving XPVM Group, cc=%d\n\n",
					cc );
			}
		}
	}

	exit( -1 );
}

